# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Record:
    def __init__(
            self, id, owner_id, acl,
            created_at=None, created_by=None,
            updated_at=None, updated_by=None, data=None):

        self._id = id
        self._owner_id = owner_id
        self._acl = acl
        self._created_at = created_at
        self._created_by = created_by
        self._updated_at = updated_at
        self._updated_by = updated_by

        self._data = data or {}

    def __len__(self):
        return len(self._data)

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value

    def __delitem__(self, key):
        del self._data[key]

    def __iter__(self):
        return iter(self._data)

    def __contains__(self, item):
        return item in self._data

    @property
    def id(self):
        return self._id

    @property
    def owner_id(self):
        return self._owner_id

    @property
    def acl(self):
        return self._acl

    @property
    def created_at(self):
        return self._created_at

    @property
    def created_by(self):
        return self._created_by

    @property
    def updated_at(self):
        return self._updated_at

    @property
    def updated_by(self):
        return self._updated_by

    @property
    def data(self):
        return self._data

    def get(self, key, default=None):
        return self._data.get(key, default)


# RecordID is immutable. Developer is not expected to modify a record's id
# once instantiated
class RecordID:
    def __init__(self, type_, key):
        if not type_:
            raise ValueError('RecordID.type cannot be None or empty')
        if not key:
            raise ValueError('RecordID.key cannot be None or empty')
        self._type = type_
        self._key = key

    @property
    def type(self):
        return self._type

    @property
    def key(self):
        return self._key


ACCESS_CONTROL_ENTRY_LEVEL_WRITE = 'write'
ACCESS_CONTROL_ENTRY_LEVEL_READ = 'read'


class AccessControlEntry:
    def __init__(self, level):
        self.level = level


ACCESS_CONTROL_ENTRY_RELATION_FRIEND = 'friend'
ACCESS_CONTROL_ENTRY_RELATION_FOLLOW = 'follow'


class RelationalAccessControlEntry(AccessControlEntry):
    def __init__(self, relation, level):
        super().__init__(level)
        self.relation = relation


class RoleAccessControlEntry(AccessControlEntry):
    def __init__(self, role, level):
        super().__init__(level)
        self.role = role


class DirectAccessControlEntry(AccessControlEntry):
    def __init__(self, user_id, level):
        super().__init__(level)
        self.user_id = user_id


class Asset:
    def __init__(self, name):
        self.name = name

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        if not name:
            raise ValueError('Asset.name cannot be None or empty')
        self._name = name


class Location:
    def __init__(self, lng, lat):
        self.lng = lng
        self.lat = lat


class Reference:
    def __init__(self, recordID):
        self.recordID = recordID

    @property
    def recordID(self):
        return self._recordID

    @recordID.setter
    def recordID(self, recordID):
        if recordID is None:
            raise ValueError('Reference.recordID cannot be None')
        self._recordID = recordID



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
from unittest.mock import call, patch

from ..bin import load_source_or_exit


class TestLoadSource(unittest.TestCase):
    @patch('skygear.bin.SourceFileLoader')
    def test_load_specified_success(self, mocker):
        load_source_or_exit('something.py')
        mocker.assert_called_once_with('plugin', 'something.py')

    @patch('skygear.bin.SourceFileLoader')
    def test_load_specified_not_found(self, mocker):
        mocker.side_effect = FileNotFoundError
        with self.assertRaises(SystemExit):
            load_source_or_exit('something.py')
        mocker.assert_called_once_with('plugin', 'something.py')

    @patch('skygear.bin.SourceFileLoader')
    def test_load_unspecified_success(self, mocker):
        load_source_or_exit(None)
        mocker.assert_called_once_with('plugin', '__init__.py')

    @patch('skygear.bin.SourceFileLoader')
    def test_load_unspecified_not_found(self, mocker):
        mocker.side_effect = FileNotFoundError
        with self.assertRaises(SystemExit):
            load_source_or_exit(None)
        mocker.assert_has_calls([
            call('plugin', '__init__.py'),
            call('plugin', 'plugin.py'),
            ])



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

import requests

from . import error


def send_action(url, payload):
    headers = {'Content-type': 'application/json',
               'Accept': 'application/json'}

    return requests.post(url, data=json.dumps(payload), headers=headers).json()


class SkygearContainer(object):
    endpoint = 'http://localhost:3000'
    api_key = None
    access_token = None
    user_id = None
    app_name = ''

    def __init__(self, endpoint=None, api_key=None, access_token=None,
                 user_id=None):
        if endpoint:
            self.endpoint = endpoint
        if api_key:
            self.api_key = api_key
        if user_id:
            self.user_id = user_id
        self.access_token = access_token

    def _request_url(self, action_name):
        endpoint = self.endpoint
        endpoint = endpoint[:-1] if endpoint[-1] == '/' else endpoint
        return endpoint + '/' + action_name.replace(':', '/')

    def _payload(self, action_name, params):
        payload = params.copy() if isinstance(params, dict) else {}
        payload['action'] = action_name
        if self.access_token:
            payload['access_token'] = self.access_token
        elif self.api_key:
            payload['api_key'] = self.api_key
        if self.user_id:
            payload['_user_id'] = self.user_id
        return payload

    @classmethod
    def set_default_app_name(cls, app_name):
        cls.app_name = app_name

    @classmethod
    def get_default_app_name(cls):
        return cls.app_name

    @classmethod
    def set_default_endpoint(cls, endpoint):
        cls.endpoint = endpoint

    @classmethod
    def set_default_apikey(cls, api_key):
        cls.api_key = api_key

    def send_action(self, action_name, params):
        resp = send_action(self._request_url(action_name),
                           self._payload(action_name, params))
        if 'error' in resp:
            raise error.SkygearException.from_dict(resp['error'])

        return resp



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
from unittest.mock import patch

from werkzeug.test import EnvironBuilder
from werkzeug.wrappers import Request

from .. import restful
from ..error import SkygearException


def create_request(*args, **kwargs):
    content_type = kwargs.pop('content_type', None)
    builder = EnvironBuilder(*args, **kwargs)
    if content_type:
        builder.content_type = content_type
    return Request(builder.get_environ())


def test_get_ident_none():
    req = create_request(method='POST', path='/hello/world')
    assert restful.get_ident('/hello/world', req) is None


def test_get_ident_none_with_slash():
    req = create_request(method='POST', path='/hello/world/')
    assert restful.get_ident('/hello/world', req) is None


def test_get_ident():
    req = create_request(method='POST', path='/hello/world/ident')
    assert restful.get_ident('/hello/world', req) == 'ident'


def test_get_ident_with_slash():
    req = create_request(method='POST', path='/hello/world/ident/')
    assert restful.get_ident('/hello/world', req) == 'ident'


def test_has_func():
    class HelloWorld:
        var = True

        def func():
            pass

    assert restful.has_func(HelloWorld(), 'func') is True
    assert restful.has_func(HelloWorld(), 'non-existent') is False
    assert restful.has_func(HelloWorld(), 'var') is False


class BlankRestfulResource(restful.RestfulResource):
    def index(self):
        pass

    def create(self):
        pass

    def delete(self, ident):
        pass

    def update(self, ident):
        pass

    def get(self, ident):
        pass


class TestRestfulResource(unittest.TestCase):
    def test_get_allowed_methods_create(self):
        class HelloWorld(restful.RestfulResource):
            def create(self):
                pass

        assert HelloWorld.get_allowed_methods() == ['POST']

    def test_get_allowed_methods_delete(self):
        class HelloWorld(restful.RestfulResource):
            def delete(self, ident):
                pass

        assert HelloWorld.get_allowed_methods() == ['DELETE']

    def test_get_allowed_methods_index(self):
        class HelloWorld(restful.RestfulResource):
            def index(self):
                pass

        assert HelloWorld.get_allowed_methods() == ['GET']

    def test_get_allowed_methods_get(self):
        class HelloWorld(restful.RestfulResource):
            def get(self, ident):
                pass

        assert HelloWorld.get_allowed_methods() == ['GET']

    def test_update_allowed_methods_update(self):
        class HelloWorld(restful.RestfulResource):
            def update(self, ident):
                pass

        assert HelloWorld.get_allowed_methods() == ['PUT']

    def test_get_payload(self):
        resource = BlankRestfulResource()
        resource.request = create_request(method='POST', path='/',
                                          data='{"json": "this is"}',
                                          content_type='application/json')
        assert resource.get_payload() == {'json': 'this is'}

    def test_get_payload_empty(self):
        resource = BlankRestfulResource()
        resource.request = create_request(method='POST', path='/')
        assert resource.get_payload() == {}

    @patch.object(BlankRestfulResource, 'get')
    def test_handle_request_get(self, mock):
        mock.return_value = {'json': 'this is'}
        resource = BlankRestfulResource()
        request = create_request(method='GET', path='/hello/1')
        assert resource.handle_request('/hello', request) == mock.return_value
        mock.assert_called_once_with('1')

    @patch.object(BlankRestfulResource, 'create')
    def test_handle_request_create(self, mock):
        mock.return_value = {'json': 'this is'}
        resource = BlankRestfulResource()
        request = create_request(method='POST', path='/hello')
        assert resource.handle_request('/hello', request) == mock.return_value
        mock.assert_called_once_with()

    @patch.object(BlankRestfulResource, 'update')
    def test_handle_request_update(self, mock):
        mock.return_value = {'json': 'this is'}
        resource = BlankRestfulResource()
        request = create_request(method='PUT', path='/hello/1')
        assert resource.handle_request('/hello', request) == mock.return_value
        mock.assert_called_once_with('1')

    @patch.object(BlankRestfulResource, 'index')
    def test_handle_request_index(self, mock):
        mock.return_value = {'result': {'json': 'this is'}}
        resource = BlankRestfulResource()
        request = create_request(method='GET', path='/hello')
        assert resource.handle_request('/hello', request) == mock.return_value
        mock.assert_called_once_with()

    @patch.object(BlankRestfulResource, 'delete')
    def test_handle_request_delete(self, mock):
        mock.return_value = {'json': 'this is'}
        resource = BlankRestfulResource()
        request = create_request(method='DELETE', path='/hello/1')
        assert resource.handle_request('/hello', request) == mock.return_value
        mock.assert_called_once_with('1')

    def test_handle_request_invalid(self):
        resource = BlankRestfulResource()
        request = create_request(method='WHAT', path='/hello/1')
        with self.assertRaises(SkygearException):
            resource.handle_request('/hello', request)


class SampleRestfulRecord(restful.RestfulRecord):
    record_type = 'sample'


class MockRestfulRecord(restful.RestfulRecord):
    record_type = 'sample'

    def _access_token(self):
        return 'ACCESS_TOKEN'

    def get_payload(self):
        return {'_id': 'sample/1', 'data': 'json'}

    def predicate(self):
        return []

    def query_options(self):
        return {'count': True}


class TestRestfulRecord(unittest.TestCase):
    def test_access_token_header(self):
        resource = SampleRestfulRecord()
        resource.request = create_request(headers={
            'X-Skygear-Access-Token': 'ACCESS_TOKEN',
            })
        assert resource._access_token() == 'ACCESS_TOKEN'

    def test_access_token_payload(self):
        resource = SampleRestfulRecord()
        resource.request = create_request(
            data='{"access_token": "ACCESS_TOKEN"}',
            content_type='application/json')
        assert resource._access_token() == 'ACCESS_TOKEN'

    def test_access_token_none(self):
        resource = SampleRestfulRecord()
        resource.request = create_request()
        assert resource._access_token() is None

    def test_handle_request_index(self):
        resource = MockRestfulRecord()

        with patch.object(resource, '_send_multi') as mock:
            resource.request = create_request()
            mock.return_value = [{'data': 'json'}]
            assert resource.index() == [{'data': 'json'}]
            mock.assert_called_once_with('record:query', database_id='_public',
                                         record_type='sample',
                                         predicate=[], count=True)

    def test_handle_request_create(self):
        resource = MockRestfulRecord()

        with patch.object(resource, '_send_single') as mock:
            mock.return_value = {'data': 'json'}
            assert resource.create() == {'data': 'json'}
            mock.assert_called_once_with('record:save', database_id='_public',
                                         records=[resource.get_payload()])

    def test_handle_request_get(self):
        resource = MockRestfulRecord()

        with patch.object(resource, '_send_single') as mock:
            mock.return_value = {'data': 'json'}
            assert resource.get('1') == {'data': 'json'}
            mock.assert_called_once_with('record:fetch', database_id='_public',
                                         ids=['sample/1'])

    def test_handle_request_update(self):
        resource = MockRestfulRecord()

        with patch.object(resource, '_send_single') as mock:
            mock.return_value = {'data': 'json'}
            assert resource.update('1') == {'data': 'json'}
            mock.assert_called_once_with('record:save', database_id='_public',
                                         records=[resource.get_payload()])

    def test_handle_request_delete(self):
        resource = MockRestfulRecord()

        with patch.object(resource, '_send_single') as mock:
            mock.return_value = {'data': 'json'}
            assert resource.delete('1') == {'data': 'json'}
            mock.assert_called_once_with('record:delete',
                                         database_id='_public',
                                         ids=['sample/1'])

    @patch('skygear.container.SkygearContainer.send_action')
    def test_send_multi(self, mock):
        mock.return_value = {'result': [{'data': 'json'}]}
        assert MockRestfulRecord()._send_multi('action', data='hello') \
            == {'result': [{'data': 'json'}]}
        mock.assert_called_once_with('action', {'data': 'hello'})

    @patch('skygear.container.SkygearContainer.send_action')
    def test_send_single(self, mock):
        mock.return_value = {'result': [{'data': 'json'}]}
        assert MockRestfulRecord()._send_single('action', data='hello') \
            == {'data': 'json'}
        mock.assert_called_once_with('action', {'data': 'hello'})



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import contextlib
import logging
import os
import re

import sqlalchemy as sa
from sqlalchemy import schema

from ..container import SkygearContainer

_app_name_pattern = re.compile('[.:]')
_engine = None
_metadata = None
_logger = logging.getLogger(__name__)


def quotedIdentifier(s):
    return '"%s"' % s.replace('"', '""')


def _get_engine():
    global _engine
    if _engine is None:
        db_url = os.getenv('DATABASE_URL')
        if not db_url:
            raise ValueError('empty environment variable "DATABASE_URL"')

        _engine = sa.create_engine(db_url)

    return _engine


def _get_schema_name():
    app_name = SkygearContainer.get_default_app_name()
    return 'app_' + _app_name_pattern.sub('_', app_name)


def _get_metadata():
    global _metadata
    if _metadata:
        return _metadata

    _metadata = _reflect_tables()
    return _metadata


def _reflect_tables():
    meta = schema.MetaData()

    schema_name = _get_schema_name()
    _logger.info("Reflecting database schema from postgres schema '{}'."
                 .format(schema_name))
    meta.reflect(bind=_get_engine(), schema=schema_name)
    return meta


def _search_path_sql():
    app_name = quotedIdentifier(_get_schema_name())
    return "SET search_path TO {0}, public;".format(app_name)


def _set_search_path(db):
    sql = _search_path_sql()
    db.execute(sql)


def get_table(name):
    schema_name = _get_schema_name()
    try:
        return _get_metadata().tables["{}.{}".format(schema_name, name)]
    except KeyError:
        raise Exception("No table of name '{}' exists in schema '{}'.",
                        name, schema_name)


@contextlib.contextmanager
def conn():
    sql = _search_path_sql()
    with _get_engine().begin() as conn:
        conn.execute(sql)
        yield conn



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest

from skygear.utils import context as ctx


class TestContextFunctions(unittest.TestCase):
    def tearDown(self):
        ctx.clear_contexts()

    def test_start_context(self):
        with ctx.start_context({'hello': 'world'}):
            assert ctx.current_context()['hello'] == 'world'
        assert ctx.current_context().get('hello', None) is None

    def test_push_context(self):
        ctx.push_context({'hello': 'world'})
        assert ctx.current_context()['hello'] == 'world'

    def test_pop_context(self):
        ctx.push_context({'hello': 'world'})
        ctx.pop_context()
        assert ctx.current_context().get('hello', None) is None

    def test_pop_context_top(self):
        with self.assertRaises(Exception):
            ctx.pop_context()

    def test_current_user_id(self):
        ctx.push_context({'user_id': '42'})
        assert ctx.current_user_id()



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

log = logging.getLogger(__name__)


def get_registry():
    return _registry


class Registry:
    """Registry holds a mapping of registred functions and their parameters that
    are callable by Skygear plugin system.

    Developers are not expected to create an instance directly. Instead, they
    should use the function get_registry() to the shared instance.
    """

    def __init__(self):
        self.func_map = {
            'op': {},
            'hook': {},
            'timer': {},
        }
        self.param_map = {
            'op': [],
            'handler': [],
            'hook': [],
            'timer': [],
            'provider': [],
        }
        self.handler = {}
        self.providers = {}

    def register(self, kind, name, func, *args, **kwargs):
        if kind == 'handler':
            return self._register_handler(name, func, kwargs)
        self.func_map[kind][name] = func
        if kind == 'hook':
            if kwargs['type'] is None:
                raise ValueError("type is required for hook")
            if kwargs['trigger'] is None:
                raise ValueError("trigger is required for hook")
            kwargs['name'] = name
            self.param_map['hook'].append(kwargs)
        elif kind == 'op':
            self.param_map['op'].append({
                'name': name,
                'auth_required': kwargs.get('auth_required',
                                            kwargs.get('key_required', False)),
                'user_required': kwargs.get('user_required', False),
                })
        elif kind == 'timer':
            kwargs['name'] = name
            self.param_map['timer'].append(kwargs)
        else:
            raise Exception("Unrecognized transport kind '%d'.".format(kind))

        log.debug("Registering %s:%s to skygear!!", kind, name)

    def _register_handler(self, name, func, kwargs):
        methods = kwargs.get('method', ['GET', 'POST', 'PUT'])
        if isinstance(methods, str):
            methods = [methods]
        self.param_map['handler'].append({
            'name': name,
            'methods': methods,
            'key_required': kwargs.get('key_required', False),
            'user_required': kwargs.get('user_required', False),
        })
        if name not in self.handler:
            self.handler[name] = {}
        for m in methods:
            self.handler[name][m] = func

    def register_provider(self, provider_type, provider_id, provider,
                          **kwargs):
        kwargs['type'] = provider_type
        kwargs['id'] = provider_id
        self.param_map['provider'].append(kwargs)
        self.providers[provider_id] = provider

    def func_list(self):
        return self.param_map

    def get_obj(self, kind, name):
        if kind == 'provider':
            return self.providers[name]
        else:
            return self.func_map[kind][name]

    def get_handler(self, name, method):
        return self.handler[name].get(method, None)

_registry = Registry()



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import json
import os
import unittest
from unittest.mock import ANY, patch

from ... import config as skyconfig
from ...registry import Registry
from ..common import encode_base64_json
from ..console import ConsoleTransport


def environ_dict(context, config):
    return {
            'SKYGEAR_CONTEXT': encode_base64_json(context).decode('utf-8'),
            'SKYGEAR_CONFIG': encode_base64_json(config).decode('utf-8')
            }


@patch('skygear.config.config', skyconfig.Configuration())
class TestConsoleTransport(unittest.TestCase):
    def exec(self, args, data):
        out = io.StringIO()
        data = data if isinstance(data, str) else json.dumps(data)
        transport = ConsoleTransport(args, stdin=io.StringIO(data), stdout=out,
                                     registry=Registry())
        transport.run()
        out.seek(0)
        return json.loads(out.read())

    @patch('skygear.transmitter.console.ConsoleTransport.init_info')
    def testInitInfo(self, mocker):
        mocker.return_value = {'data': 'hello'}
        output = self.exec(['init'], "")
        mocker.assert_called_once_with()
        assert output == mocker.return_value

    @patch.dict(os.environ, environ_dict({'context': 'happy'},
                                         {'hello': 'bye'}))
    @patch('skygear.transmitter.console.ConsoleTransport.call_func')
    def testCallFuncWithData(self, mocker):
        mocker.return_value = {'data': 'hello'}
        output = self.exec(['timer', 'name'], {'data': 'bye'})
        mocker.assert_called_once_with({'context': 'happy'}, 'timer', 'name',
                                       {'data': 'bye'})
        assert skyconfig.config.hello == 'bye'
        assert output == mocker.return_value

    @patch('skygear.transmitter.console.ConsoleTransport.call_func')
    def testOpInvalidJSON(self, mocker):
        mocker.return_value = {}
        output = self.exec(['op', 'john'], "not-valid")
        assert not mocker.called
        expected_output = {'code': 10000, 'info': {},
                           'message': 'unable to parse JSON string'}
        assert output == expected_output

    @patch('skygear.transmitter.console.ConsoleTransport.call_func')
    def testOpEmpty(self, mocker):
        mocker.return_value = {"data": "hi"}
        output = self.exec(['op', 'john'], "")
        mocker.assert_called_once_with(ANY, 'op', 'john', {})
        assert output == mocker.return_value

    @patch('skygear.transmitter.console.ConsoleTransport.call_func')
    def testOp(self, mocker):
        mocker.return_value = {"data": "hi"}
        output = self.exec(['op', 'john'], {'data': 'bye'})
        mocker.assert_called_once_with(ANY, 'op', 'john', {'data': 'bye'})
        assert output == mocker.return_value

    @patch('skygear.transmitter.console.ConsoleTransport.call_func')
    def testHook(self, mocker):
        mocker.return_value = {}
        self.exec(['hook', 'john'], {'data': 'bye'})
        mocker.assert_called_once_with(ANY, 'hook', 'john', {'data': 'bye'})

    @patch('skygear.transmitter.console.ConsoleTransport.call_provider')
    def testProvider(self, mocker):
        mocker.return_value = {}
        self.exec(['provider', 'john', 'work'], {'data': 'bye'})
        mocker.assert_called_once_with(ANY, 'john', 'work', {'data': 'bye'})

    @patch('skygear.transmitter.console.ConsoleTransport.call_func')
    def testTimer(self, mocker):
        mocker.return_value = {}
        self.exec(['timer', 'john'], "")
        mocker.assert_called_once_with(ANY, 'timer', 'john', {})



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ..container import SkygearContainer


class TestContainer():
    def test_payload_include_action(self):
        c = SkygearContainer(endpoint='endpoint', access_token='access-token')
        payload = c._payload('action:work', {'data': 'hello world'})
        assert payload['action'] == 'action:work'
        assert payload['data'] == 'hello world'

    def test_payload_include_api_key(self):
        c = SkygearContainer(endpoint='endpoint', api_key='api-key')
        payload = c._payload('action:work', {})
        assert payload['api_key'] == 'api-key'

    def test_payload_include_user_id(self):
        c = SkygearContainer(endpoint='endpoint', api_key='api-key',
                             user_id='user-id')
        payload = c._payload('action:work', {})
        assert payload['_user_id'] == 'user-id'

    def test_payload_include_access_token(self):
        c = SkygearContainer(endpoint='endpoint', access_token='access-token')
        payload = c._payload('action:work', {})
        assert payload['access_token'] == 'access-token'



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime

from .registry import get_registry

_registry = get_registry()


def op(name, *args, **kwargs):
    def our_op(func):
        _registry.register("op", name, func, *args, **kwargs)
        return func
    return our_op


def every(interval, name=None, *args, **kwargs):
    if isinstance(interval, datetime.timedelta):
        interval = "@every {0}s".format(interval.total_seconds())
    elif isinstance(interval, int):
        interval = "@every {0}s".format(interval)
    elif not isinstance(interval, str):
        msg = "Expecting int, timedelta or str for interval. Got '{0}'." \
            .format(type(interval).__name__)
        raise Exception(msg)
    kwargs['spec'] = interval
    kwargs['name'] = name

    def our_every(func):
        name = kwargs.pop('name', None) or \
            func.__module__ + "." + func.__name__

        _registry.register("timer", name, func, *args, **kwargs)
        return func
    return our_every


def handler(name, *args, **kwargs):
    def skygear_handler(func):
        _registry.register("handler", name, func, *args, **kwargs)
        return func
    return skygear_handler


def hook(trigger, *args, **kwargs):
    kwargs['trigger'] = trigger

    def skygear_hook(func):
        def hook_func(record, original_record, db):
            # return the record for user
            func(record, original_record, db)
            return record

        name = kwargs.pop('name', None) or \
            func.__module__ + "." + func.__name__

        _registry.register("hook", name, hook_func, *args, **kwargs)
        return func

    return skygear_hook


def register_save_hook(func, *args, **kwargs):
    def hook_func(record, original_record, db):
        func(record, original_record, db)
        return record

    name = kwargs.pop('name', None) or \
        func.__module__ + "." + func.__name__

    _registry.register("hook", name, hook_func, *args, **kwargs)


def register_delete_hook(func, *args, **kwargs):
    def hook_func(record, original_record, db):
        func(record, db)
        return record

    name = kwargs.pop('name', None) or \
        func.__module__ + "." + func.__name__

    _registry.register("hook", name, hook_func, *args, **kwargs)


def before_save(type_, *args, **kwargs):
    kwargs['type'] = type_
    kwargs['trigger'] = "beforeSave"

    def wrapper(func):
        register_save_hook(func, *args, **kwargs)
        return func
    return wrapper


def after_save(type_, *args, **kwargs):
    kwargs['type'] = type_
    kwargs['trigger'] = "afterSave"

    def wrapper(func):
        register_save_hook(func, *args, **kwargs)
        return func
    return wrapper


def before_delete(type_, *args, **kwargs):
    kwargs['type'] = type_
    kwargs['trigger'] = "beforeDelete"

    def wrapper(func):
        register_delete_hook(func, *args, **kwargs)
        return func
    return wrapper


def after_delete(type_, *args, **kwargs):
    kwargs['type'] = type_
    kwargs['trigger'] = "afterDelete"

    def wrapper(func):
        register_delete_hook(func, *args, **kwargs)
        return func
    return wrapper


def provides(provider_type, provider_id, *args, **kwargs):
    def skygear_provider(klass):
        provider = klass()
        _registry.register_provider(provider_type, provider_id, provider,
                                    *args, **kwargs)
        return klass
    return skygear_provider


def _fix_handler_path(name):
    name = name.replace(':', '/')
    if name.startswith('/'):
        name = name[1:]
    if name.endswith('/'):
        name = name[:-1]
    return name


def rest(name, *args, **kwargs):
    name = _fix_handler_path(name)

    def wrapper(klass):
        if callable(getattr(klass, 'get_allowed_methods', None)):
            kwargs['method'] = klass.get_allowed_methods()
        else:
            kwargs['method'] = ['GET', 'POST', 'PUT', 'DELETE']

        base_name = name if name.startswith('/') else '/' + name

        @handler(name, *args, **kwargs)
        @handler(name + '/', *args, **kwargs)
        def call_restful_wraps(request):
            restful = klass()
            return restful.handle_request(base_name, request)
        return klass
    return wrapper



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import signal
import sys
from importlib.machinery import SourceFileLoader

from .container import SkygearContainer
from .options import parse_args
from .transmitter import ConsoleTransport, HttpTransport, ZmqTransport

log = logging.getLogger(__name__)


def main():
    options = parse_args()
    setup_logging(options)
    run_plugin(options)


def load_source_or_exit(source):
    """
    Load the specified source file.

    If the source file evaluates to False, this function will attempt
    to load __init__.py followed by plugin.py, the first found source
    will be loaded.

    If no source is found, this function will exit the program.
    """
    if source:
        try:
            SourceFileLoader('plugin', source).load_module()
        except FileNotFoundError:
            log.error("File not found: {0}".format(source))
            sys.exit(1)
    else:
        for x in ['__init__.py', 'plugin.py']:
            try:
                SourceFileLoader('plugin', x).load_module()
                break
            except FileNotFoundError:
                pass
        else:
            log.error("Unable to find __init__.py or plugin.py.")
            sys.exit(1)


def run_plugin(options):
    load_source_or_exit(options.plugin)

    SkygearContainer.set_default_app_name(options.appname)
    SkygearContainer.set_default_endpoint(options.skygear_endpoint)
    SkygearContainer.set_default_apikey(options.apikey)

    log.debug("Install signal handler for SIGTERM")
    signal.signal(signal.SIGTERM, sigterm_handler)

    if options.subprocess is not None:
        transport = ConsoleTransport(options.subprocess)
    elif options.http:
        transport = HttpTransport(options.http_addr, debug=options.debug)
    else:
        log.info(
            "Connecting to address %s" % options.skygear_address)
        transport = ZmqTransport(options.skygear_address)
    transport.run()


def setup_logging(options):
    # TODO: Make it load a stadard python logging config.
    logger = logging.getLogger()
    level = {
        'DEBUG': logging.DEBUG,
        'INFO': logging.INFO,
        'WARN': logging.WARN,
        'ERROR': logging.ERROR,
        'CRITICAL': logging.CRITICAL
    }.get(options.loglevel, logging.INFO)
    logger.setLevel(level)
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('''\
%(asctime)s %(levelname)-5.5s \
[%(name)s:%(lineno)s][%(threadName)s] %(message)s\
''')
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def sigterm_handler(signum, frame):
    """
    When docker stops the container, it sends PID 1 with SIGTERM.
    This function exits the program when SIGTERM is received.

    Remember not to run this program in a shell as PID 1 because the shell may
    not forward the signal.
    """
    sys.exit(1)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import traceback
from datetime import datetime

import strict_rfc3339

from ..error import SkygearException, UnexpectedError
from ..models import (Asset, DirectAccessControlEntry, Location, Record,
                      RecordID, Reference, RelationalAccessControlEntry,
                      RoleAccessControlEntry)


def _serialize_exc(e):
    return SkygearException(
        str(e),
        UnexpectedError,
        {'trace': traceback.format_exc()})


def deserialize_record(obj):
    return _RecordDecoder().decode(obj)


def deserialize_or_none(obj):
    if obj:
        return deserialize_record(obj)
    else:
        return None


def serialize_record(record):
    return _RecordEncoder().encode(record)


class _RecordDecoder:
    def decode(self, d):
        id = self.decode_id(d['_id'])
        owner_id = d['_ownerID']
        acl = self.decode_acl(d['_access'])

        created_at = None
        if d.get('_created_at', None):
            created_at = self.decode_date_value(d['_created_at'])
        created_by = d.get('_created_by', None)

        updated_at = None
        if d.get('_updated_at', None):
            updated_at = self.decode_date_value(d['_updated_at'])
        updated_by = d.get('_updated_by', None)

        data_dict = {k: v for k, v in d.items() if not k.startswith('_')}

        data = self.decode_dict(data_dict)

        return Record(
            id=id,
            owner_id=owner_id,
            acl=acl,
            created_at=created_at,
            created_by=created_by,
            updated_at=updated_at,
            updated_by=updated_by,
            data=data)

    def decode_id(self, s):
        ss = s.split('/')
        return RecordID(ss[0], ss[1])

    def decode_acl(self, l):
        if l is None:
            return None
        if not isinstance(l, list):
            raise TypeError('expect ACL to be a list')

        return [self.decode_ace(d) for d in l]

    def decode_ace(self, d):
        level = d.get('level', None)
        if level is None:
            raise ValueError("ace must have level")
        relation = d.get('relation', None)
        role = d.get('role', None)
        user_id = d.get('user_id', None)
        if user_id is not None:
            return DirectAccessControlEntry(d['user_id'], level)
        elif relation is not None:
            return RelationalAccessControlEntry(relation, level)
        elif role is not None:
            return RoleAccessControlEntry(role, level)
        else:
            raise ValueError("invalid ace")

    def decode_dict(self, d):
        return {k: self.decode_value(v) for k, v in d.items()}

    def decode_list(self, l):
        return [self.decode_value(v) for v in l]

    def decode_value(self, v):
        if isinstance(v, dict):
            type_ = v.get('$type')
            if type_ == 'date':
                return self.decode_date(v)
            elif type_ == 'asset':
                return self.decode_asset(v)
            elif type_ == 'geo':
                return self.decode_location(v)
            elif type_ == 'ref':
                return self.decode_ref(v)
            else:
                return self.decode_dict(v)
        elif isinstance(v, list):
            return self.decode_list(v)
        else:
            return v

    def decode_date(self, d):
        return self.decode_date_value(d['$date'])

    def decode_date_value(self, s):
        ts = strict_rfc3339.rfc3339_to_timestamp(s)
        return datetime.fromtimestamp(ts)

    def decode_asset(self, d):
        return Asset(d['$name'])

    def decode_location(self, d):
        return Location(d['$lng'], d['$lat'])

    def decode_ref(self, d):
        return Reference(self.decode_id(d['$id']))


class _RecordEncoder:
    def encode(self, record):
        d = self.encode_dict(record.data)
        d['_id'] = self.encode_id(record.id)
        d['_ownerID'] = record.owner_id
        d['_access'] = self.encode_acl(record.acl)
        return d

    def encode_id(self, id):
        return '%s/%s' % (id.type, id.key)

    def encode_acl(self, acl):
        if acl is None:
            return None

        return [self.encode_ace(e) for e in acl]

    def encode_ace(self, ace):
        if isinstance(ace, RelationalAccessControlEntry):
            return {
                'level': ace.level,
                'relation': ace.relation,
            }
        elif isinstance(ace, DirectAccessControlEntry):
            return {
                'level': ace.level,
                'user_id': ace.user_id,
            }
        elif isinstance(ace, RoleAccessControlEntry):
            return {
                'level': ace.level,
                'role': ace.role
            }
        else:
            raise ValueError('Unknown type of ACE = %s', type(ace))

    def encode_dict(self, d):
        return {k: self.encode_value(v) for k, v in d.items()}

    def encode_list(self, l):
        return [self.encode_value(v) for v in l]

    def encode_value(self, v):
        if isinstance(v, dict):
            return self.encode_dict(v)
        elif isinstance(v, list):
            return self.encode_list(v)
        elif isinstance(v, datetime):
            return self.encode_datetime(v)
        elif isinstance(v, Asset):
            return self.encode_asset(v)
        elif isinstance(v, Location):
            return self.encode_location(v)
        elif isinstance(v, Reference):
            return self.encode_ref(v)
        else:
            return v

    def encode_datetime(self, dt):
        ts = dt.timestamp()
        return {
            '$type': 'date',
            '$date': strict_rfc3339.timestamp_to_rfc3339_utcoffset(ts),
        }

    def encode_asset(self, asset):
        return {
            '$type': 'asset',
            '$name': asset.name,
        }

    def encode_location(self, location):
        return {
            '$type': 'geo',
            '$lng': location.lng,
            '$lat': location.lat,
        }

    def encode_ref(self, ref):
        return {
            '$type': 'ref',
            '$id': self.encode_id(ref.recordID),
        }



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import contextlib
import logging
import threading

log = logging.getLogger(__name__)


class RequestContextManager(threading.local):
    def __init__(self):
        self.clear()

    def push(self, context):
        self.stack.append(context)
        log.debug("Pushed request context: %s. Length: %d",
                  context, len(self.stack))

    def pop(self):
        if len(self.stack) == 1:
            raise Exception("Cannot pop the top-most context.")
        return self.stack.pop()

    def current(self):
        return self.stack[-1]

    def clear(self):
        self.stack = [{}]


_manager = RequestContextManager()


def _context_manager():
    return _manager


def push_context(context):
    _context_manager().push(context)


def current_context():
    return _context_manager().current()


def pop_context():
    return _context_manager().pop()


def clear_contexts():
    _context_manager().clear()


def current_user_id():
    """Return the current user ID of the current request context.

    If the user is not logged in or if the request context is not available.
    This function returns None.
    """
    return current_context().get("user_id", None)


@contextlib.contextmanager
def start_context(context):
    push_context(context)
    try:
        yield
    finally:
        pop_context()



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging

from skygear.container import SkygearContainer
from skygear.error import BadRequest, SkygearException, UnexpectedError


log = logging.getLogger(__name__)


def get_ident(base_name, request):
    if not request.path.startswith(base_name):
        raise Exception('path "{}" not matching "{}"'
                        .format(request.path, base_name))
    path = request.path[len(base_name):]
    if path.startswith('/'):
        path = path[1:]
    return path.split('/')[0] or None


def has_func(obj, name):
    return callable(getattr(obj, name, None))


class RestfulResource:
    @classmethod
    def get_allowed_methods(cls):
        methods = []
        for r in cls._routes():
            if has_func(cls, r['func']):
                methods.append(r['method'])
        return list(set(methods))

    def get_payload(self):
        data = self.request.get_data(as_text=True)
        if data == '':
            return {}
        try:
            payload = json.loads(data)
            return payload
        except ValueError:
            raise SkygearException('unable to decode json', BadRequest)

    @classmethod
    def _routes(cls):
        return [
                {'ident': True, 'method': 'PUT', 'func': 'update'},
                {'ident': True, 'method': 'GET', 'func': 'get'},
                {'ident': True, 'method': 'DELETE', 'func': 'delete'},
                {'ident': False, 'method': 'POST', 'func': 'create'},
                {'ident': False, 'method': 'GET', 'func': 'index'},
                ]

    def handle_request(self, base_name, request):
        self.request = request
        ident = get_ident(base_name, request)

        for r in self._routes():
            if r['ident'] != bool(ident):
                continue
            if r['method'] != request.method:
                continue
            if not has_func(self, r['func']):
                continue

            if ident:
                return getattr(self, r['func'])(ident)
            else:
                return getattr(self, r['func'])()
        else:
            raise SkygearException('invalid request method', BadRequest)


class RestfulRecord(RestfulResource):
    record_type = None
    database_id = '_public'

    def _send_multi(self, action, **payload):
        token = self._access_token()
        container = SkygearContainer(access_token=token)
        result = container.send_action(action, payload)
        if 'error' in result:
            raise SkygearException.from_dict(result['error'])
        elif 'result' in result and isinstance(result['result'], list):
            return result
        else:
            raise SkygearException('unexpected result', UnexpectedError)

    def _send_single(self, action, **payload):
        token = self._access_token()
        container = SkygearContainer(access_token=token)
        result = container.send_action(action, payload)
        if 'error' in result:
            raise SkygearException.from_dict(result['error'])
        elif 'result' in result and isinstance(result['result'], list) \
                and len(result['result']) > 0:
            first_result = result['result'][0]
            if first_result.get('_type', None) == 'error':
                raise SkygearException.from_dict(first_result)
            return first_result
        else:
            raise SkygearException('unexpected result', UnexpectedError)

    def _record_id(self, ident):
        return self.record_type + '/' + ident

    def _access_token(self):
        token = self.request.headers.get('X-Skygear-Access-Token', None)
        if token:
            return token

        try:
            data = json.loads(self.request.get_data(as_text=True))
            return data.get('access_token', None)
        except ValueError:
            return None

    def predicate(self):
        return None

    def query_options(self):
        return {}

    def index(self):
        """
        List records by querying the database.
        """
        options = self.query_options() or {}
        predicate = self.predicate()
        if predicate is not None:
            options['predicate'] = predicate

        return self._send_multi('record:query',
                                database_id=self.database_id,
                                record_type=self.record_type,
                                **options
                                )

    def create(self):
        payload = self.get_payload()
        if not payload.get('_id', '').startswith(self.record_type + '/'):
            raise SkygearException()
        return self._send_single('record:save',
                                 database_id=self.database_id,
                                 records=[payload])

    def delete(self, ident):
        record_id = self._record_id(ident)
        return self._send_single('record:delete',
                                 database_id=self.database_id,
                                 ids=[record_id])

    def update(self, ident):
        payload = self.get_payload()
        payload['_id'] = self._record_id(ident)
        return self._send_single('record:save',
                                 database_id=self.database_id,
                                 records=[payload])

    def get(self, ident):
        record_id = self._record_id(ident)
        return self._send_single('record:fetch',
                                 database_id=self.database_id,
                                 ids=[record_id])



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest

from ..registry import Registry


class TestRegistry(unittest.TestCase):
    def test_register_handler(self):
        def handler():
            pass

        kwargs = {
            'method': ['GET', 'POST'],
            'key_required': True,
            'user_required': True,
        }
        registry = Registry()
        registry.register('handler', 'plugin:handler', handler, **kwargs)

        assert len(registry.handler) == 1
        assert registry.get_handler('plugin:handler', 'GET') == handler
        assert registry.get_handler('plugin:handler', 'POST') == handler
        assert registry.get_handler('plugin:handler', 'PUT') is None

        param_map = registry.param_map['handler']
        assert len(param_map) == 1
        assert param_map[0]['name'] == 'plugin:handler'
        assert 'POST' in param_map[0]['methods']
        assert param_map[0]['key_required'] is True
        assert param_map[0]['user_required'] is True

    def test_register_handler_with_one_str_method(self):
        def handler():
            pass

        kwargs = {
            'method': 'PUT',
        }
        registry = Registry()
        registry.register('handler', 'plugin:handler', handler, **kwargs)

        assert len(registry.handler) == 1
        assert registry.get_handler('plugin:handler', 'GET') is None
        assert registry.get_handler('plugin:handler', 'POST') is None
        assert registry.get_handler('plugin:handler', 'PUT') == handler

        param_map = registry.param_map['handler']
        assert len(param_map) == 1
        assert param_map[0]['name'] == 'plugin:handler'
        assert 'PUT' in param_map[0]['methods']
        assert param_map[0]['key_required'] is False
        assert param_map[0]['user_required'] is False

    def test_register_op(self):
        def fn():
            pass

        kwargs = {
                'key_required': True,
                'user_required': True,
                }
        registry = Registry()
        registry.register('op', 'plugin:action', fn, **kwargs)

        func_map = registry.func_map['op']
        assert len(func_map) == 1
        assert func_map['plugin:action'] == fn

        param_map = registry.param_map['op']
        assert len(param_map) == 1
        assert param_map[0]['name'] == 'plugin:action'
        assert param_map[0]['auth_required'] is True
        assert param_map[0]['user_required'] is True

    def test_register_hook(self):
        def fn():
            pass

        kwargs = {
                'type': 'note',
                'trigger': 'beforeSave',
                }
        registry = Registry()
        registry.register('hook', 'hook_name', fn, **kwargs)

        func_map = registry.func_map['hook']
        assert len(func_map) == 1
        assert func_map['hook_name'] == fn

        param_map = registry.param_map['hook']
        assert len(param_map) == 1
        assert param_map[0]['name'] == 'hook_name'
        assert param_map[0]['type'] == 'note'
        assert param_map[0]['trigger'] == 'beforeSave'

    def test_register_timer(self):
        def fn():
            pass

        registry = Registry()
        registry.register('timer', 'timer_name', fn)

        func_map = registry.func_map['timer']
        assert len(func_map) == 1
        assert func_map['timer_name'] == fn

        param_map = registry.param_map['timer']
        assert len(param_map) == 1
        assert param_map[0]['name'] == 'timer_name'

    def test_register_provider(self):
        def fn():
            pass

        class Provider():
            pass

        the_provider = Provider()
        registry = Registry()
        registry.register_provider('example', 'com.example', the_provider)

        assert len(registry.providers) == 1
        assert registry.providers['com.example'] == the_provider

        param_map = registry.param_map['provider']
        assert len(param_map) == 1
        assert param_map[0]['type'] == 'example'
        assert param_map[0]['id'] == 'com.example'



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import unittest
from unittest.mock import ANY, patch

from werkzeug.test import Client
from werkzeug.wrappers import BaseResponse

from ... import config as skyconfig
from ...registry import Registry
from ..common import encode_base64_json
from ..http import HttpTransport


def headers_with_context(data):
    return {'X-Skygear-Plugin-Context': encode_base64_json(data)}


@patch('skygear.config.config', skyconfig.Configuration())
class TestHttpTransport(unittest.TestCase):
    def get_app(self):
        return self.transport.dispatch

    def get_client(self):
        return Client(self.get_app(), BaseResponse)

    def setUp(self):
        self.transport = HttpTransport('127.0.0.1:8888', Registry())

    def tearDown(self):
        self.transport = None

    @patch('skygear.transmitter.http.run_simple')
    def testRun(self, mocker):
        self.transport.run()
        mocker.assert_called_once_with('127.0.0.1', 8888,
                                       self.transport.dispatch,
                                       use_reloader=False)

    @patch('skygear.transmitter.http.HttpTransport.init_info')
    def testInitInfo(self, mocker):
        mocker.return_value = {'data': 'hello'}
        request_data = json.dumps({'config': {'hello': 'world'}})
        resp = self.get_client().post('/init', data=request_data)
        assert resp.status_code == 200
        mocker.assert_called_once_with()
        assert json.loads(resp.get_data(as_text=True)) == mocker.return_value
        assert skyconfig.config.hello == 'world'

    @patch('skygear.transmitter.http.HttpTransport.call_func')
    def testCallFuncWithData(self, mocker):
        mocker.return_value = {'data': 'hello'}
        request_data = json.dumps({'data': 'bye'})
        headers = headers_with_context({'context': 'happy'})
        resp = self.get_client().post('/timer/name', data=request_data,
                                      headers=headers)
        assert resp.status_code == 200
        mocker.assert_called_once_with({'context': 'happy'}, 'timer', 'name',
                                       {'data': 'bye'})
        assert json.loads(resp.get_data(as_text=True)) == mocker.return_value

    @patch('skygear.transmitter.http.HttpTransport.call_func')
    def testOp(self, mocker):
        mocker.return_value = {}
        resp = self.get_client().post('/op/john')
        assert resp.status_code == 200
        mocker.assert_called_once_with(ANY, 'op', 'john', ANY)

    @patch('skygear.transmitter.http.HttpTransport.call_func')
    def testHook(self, mocker):
        mocker.return_value = {}
        resp = self.get_client().post('/hook/john')
        assert resp.status_code == 200
        mocker.assert_called_once_with(ANY, 'hook', 'john', ANY)

    @patch('skygear.transmitter.http.HttpTransport.call_provider')
    def testProvider(self, mocker):
        mocker.return_value = {}
        resp = self.get_client().post('/provider/john/work')
        assert resp.status_code == 200
        mocker.assert_called_once_with(ANY, 'john', 'work', ANY)
        args, kwargs = mocker.call_args

    @patch('skygear.transmitter.http.HttpTransport.call_func')
    def testTimer(self, mocker):
        mocker.return_value = {}
        resp = self.get_client().post('/timer/john')
        assert resp.status_code == 200
        mocker.assert_called_once_with(ANY, 'timer', 'john', ANY)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import bcrypt
from sqlalchemy.sql import text

from .db import conn


def hash_password(password):
    """
    Return a hased password of a plaintext password.
    """
    hashed = bcrypt.hashpw(password.encode('utf-8'), bcrypt.gensalt())
    return hashed.decode("utf-8")


def reset_password_by_username(username, new_password):
    """
    Reset the user password with a new password.
    """
    if not (isinstance(username, str) and isinstance(new_password, str)):
        raise ValueError("username and new_password must be string")

    sql = text('''
        UPDATE \"_user\"
        SET password = :new_password
        WHERE username = :username
        ''')
    with conn() as db:
        result = db.execute(sql,
                            new_password=hash_password(new_password),
                            username=username)
    return result.rowcount > 0
